
import { Req } from '../http/request';
import { UserData } from '../storage';
import { HttpConfig } from '../http/server';
import { HttpWebDependencies } from './server';
import { csp_headers } from '../http/content-security-policy';
import { FastifyInstance, RouteShorthandOptions } from 'fastify';
import { render_template } from '../utilities/mustache';
import type { NamedLocation } from '../storage/named-location';

export function register_root_page_endpoint(http_server: FastifyInstance, conf: HttpConfig, { session, services }: HttpWebDependencies) {
	const opts: RouteShorthandOptions = {
		schema: { },
	};

	http_server.get('/', opts, async (req: Req, res) => {
		try {
			await session.check_login(req);
		}

		catch (error) {
			session.reset(res);
		}

		const location: NamedLocation = {
			name: 'Home',
			latitude: 45.4961,
			longitude: -122.6714,
		};

		const [ forecast, alerts, weather ] = await Promise.all([
			services.weather_gov.get_forecast_for_location(location, 'us'),
			services.weather_gov.get_alerts_for_location(location),
			services.weatherapi_com.get_current_for_location(location),
		]);

		const rendered_widgets = await Promise.all([
			services.weatherapi_com.render_current_weather(location, weather),
			services.weather_gov.render_forecast(location, forecast, alerts),
		]);

		res.status(200);
		res.header('cache-control', 'private, no-cache');
		res.header('content-type', 'text/html; charset=utf-8');
		csp_headers(res, conf.exposed_url);
		return render_root_page(req.session?.user, rendered_widgets);
	});
}

function render_root_page(user: UserData, rendered_widgets: string[]) {
	const view = {
		user,
		page_title: 'Dashboard',
		rendered_widgets,
	};

	const partial_files = {
		controls: 'controls.html.mustache',
		color_theme_controls: 'color-theme-controls.html.mustache',
		page_content: 'root.html.mustache',
	};

	return render_template('page.html.mustache', view, { }, partial_files);
}
