
import pino from 'pino';
import { URL } from 'url';
import { request as http_request, OutgoingHttpHeaders, IncomingHttpHeaders } from 'http';
import { request as https_request } from 'https';
import { create_request_cache_provider } from './request-cache';
import { HttpURL } from '../utilities/types';
import { deep_copy } from '../utilities/deep';

let next_id = 1;
const timeout = 5000;

export interface HttpResult {
	url: URL;
	// req: ClientRequest;
	// res: IncomingMessage;
	status: number;
	body: string;
	headers: IncomingHttpHeaders;
}

export interface OutboundHttpConfig {
	https_only: boolean;
}

export function validate_outbound_http_conf(conf: unknown) : asserts conf is OutboundHttpConfig {
	// todo: validate config
}

export type OutboundHttpProvider = ReturnType<typeof create_outbound_http_provider>;

export function create_outbound_http_provider(conf: OutboundHttpConfig, logger: pino.Logger) {
	const http_log = logger.child({ logger: 'outbound-http' });
	const response_cache = create_request_cache_provider({ }, http_log);

	return {	
		req: http_req,	
		get(url_str: HttpURL, headers?: OutgoingHttpHeaders, skip_cache = false) {
			return http_req('GET', url_str, headers, null, skip_cache);
		},
		head(url_str: HttpURL, headers?: OutgoingHttpHeaders, skip_cache = false) {
			return http_req('HEAD', url_str, headers, null, skip_cache);
		},
		post(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('POST', url_str, headers, body, skip_cache);
		},
		put(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('PUT', url_str, headers, body, skip_cache);
		},
		patch(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('PATCH', url_str, headers, body, skip_cache);
		},
		delete(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('DELETE', url_str, headers, body, skip_cache);
		},
	};

	async function http_req(method: string, url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
		const url = new URL(url_str);
		const make_request
			= url.protocol === 'https:' ? https_request
			: url.protocol === 'http:' ? http_request
			: null;
	
		if (! make_request) {
			throw new Error('illegal protocol');
		}
	
		if (url.protocol !== 'https:' && conf.https_only) {
			throw new Error('non-secure http requests not allowed (by configuration)');
		}
	
		if (url.username || url.password) {
			throw new Error('urls containing user credentials not allowed');
		}

		const req_headers = headers ? deep_copy(headers) : { };
	
		const result: HttpResult = {
			url,
			// req: null,
			// res: null,
			status: null,
			body: null,
			headers: null,
		};
	
		const path = url.pathname + (url.search || '');
		const port = url.port ? parseInt(url.port, 10) : (url.protocol === 'https:' ? 443 : 80);
		const log = http_log.child({
			req_id: next_id++,
			method: method,
			protocol: url.protocol,
			hostname: url.hostname,
			port: port,
			path: path,
		});

		const cached = skip_cache
			? null
			: response_cache.check_for_cached_response(method, url_str, req_headers);

		if (cached) {
			const revalidation = response_cache.determine_revalidation_needed(cached);

			if (! revalidation.should_revalidate) {
				log.info('serving request from cache (no revalidation required)');
				result.status = cached.status;
				result.body = cached.body;
				result.headers = deep_copy(cached.res_headers);
				return Promise.resolve(result);
			}

			// if (! revalidation.must_revalidate) {
			// 	// todo: lazy revalidation option
			// }

			log.info('cached response found, but requires revalidation');
			Object.assign(req_headers, revalidation.headers);
		}
	
		log.info('starting outbound http request');
	
		return new Promise<HttpResult>((resolve, reject) => {
			const start = Date.now();
			const req = make_request({
				method: method,
				protocol: url.protocol,
				hostname: url.hostname,
				port: port,
				path: path,
				headers: req_headers
			}, (res) => {
				// result.res = res;
				result.status = res.statusCode;

				log.info({
					status: res.statusCode,
					duration: Date.now() - start,
				}, `headers received`);

				switch (result.status) {
					case 304:
						log.info({ status: cached.status }, 're-using cached response (not modified)');
						result.status = cached.status;
						result.body = cached.body;
						result.headers = deep_copy(cached.res_headers);
						response_cache.freshen_cache(cached, res.headers);
						resolve(result);
						break;

					case 500:
					case 502:
					case 503:
					case 504:
						if (cached?.cache_info?.cache_control?.stale_if_error) {
							log.info({ status: cached.status }, 're-using cached response (response returned an error status, and `stale-if-error` directive was provided)');
							result.status = cached.status;
							result.body = cached.body;
							result.headers = deep_copy(cached.res_headers);
							resolve(result);
							break;
						}
						// nobreak

					default: {
						const start = Date.now();
						let data = '';
			
						res.on('data', (chunk) => {
							data += chunk;
						});
			
						res.on('end', () => {
							log.info({ duration: Date.now() - start }, `response finished`);
							result.body = data;
							result.headers = res.headers;
							response_cache.add_response_to_cache(method, url_str, req_headers, result);
							resolve(result);
						});
					};
				}
			});
	
			// result.req = req;
		
			req.setTimeout(timeout, () => {
				log.error('request timeout');
				req.destroy();
			});
		
			req.on('error', (error) => {
				log.error('request error ' + error.stack);
				reject(error);
			});
	
			if (body) {
				req.write(body);
			}
	
			req.end();
		});
	}
}
