
import pino from 'pino';
import { deep_freeze } from '../../utilities/deep';
import { render_forecast } from './render-forecast';
import type { NamedLocation } from '../../storage/named-location';
import type { OutboundHttpProvider } from '../../http/outbound';
import type { WeatherGovPoints, WeatherGovForecast, WeatherGovAlerts } from './interface';
import { HttpURL } from '../../utilities/types';

export interface WeatherGovConfig {
	enabled: boolean;
	user_agent: string;
}

export function validate_weather_gov_conf(conf: unknown) : asserts conf is WeatherGovConfig {
	// todo: validate config
}

export type WeatherGovProvider = ReturnType<typeof create_weather_gov_provider>;

export function create_weather_gov_provider(conf: WeatherGovConfig, logger: pino.Logger, outbound_http: OutboundHttpProvider) {
	const log = logger.child({ logger: 'weather.gov' });

	return {
		ready: Promise.resolve(),

		get enabled() {
			return conf?.enabled;
		},

		get_points: if_enabled(get_points),
		get_forecast: if_enabled(get_forecast),
		get_alerts: if_enabled(get_alerts),

		async get_forecast_for_lat_long(latitude: number, longitude: number, units: 'us' | 'si') {
			const points = await get_points(latitude, longitude);
			return get_forecast(points.gridId, points.gridX, points.gridY, units, false);
		},

		async get_hourly_forecast_for_lat_long(latitude: number, longitude: number, units: 'us' | 'si') {
			const points = await get_points(latitude, longitude);
			return get_forecast(points.gridId, points.gridX, points.gridY, units, true);
		},

		async get_forecast_for_location(location: NamedLocation, units: 'us' | 'si') {
			const points = await get_points(location.latitude, location.longitude);
			return get_forecast(points.gridId, points.gridX, points.gridY, units, false);
		},

		async get_hourly_forecast_for_location(location: NamedLocation, units: 'us' | 'si') {
			const points = await get_points(location.latitude, location.longitude);
			return get_forecast(points.gridId, points.gridX, points.gridY, units, true);
		},

		async get_alerts_for_location(location: NamedLocation) {
			return get_alerts(location.latitude, location.longitude);
		},

		render_forecast,
	};

	function if_enabled<T>(func: T) : T {
		if (conf?.enabled) {
			return func;
		}

		return (() => {
			throw new Error('weather.gov service not enabled');
		}) as T;
	}
	
	async function get_points(latitude: number, longitude: number) : Promise<WeatherGovPoints> {
		log.info({ latitude, longitude }, 'fetching points data');

		let { status, headers, body } = await outbound_http.get(`https://api.weather.gov/points/${latitude},${longitude}`, {
			'accept': 'application/ld+json',
			'user-agent': conf.user_agent,
		});

		// Follow a single redirect if given
		if (status === 301) {
			({ status, headers, body } = await outbound_http.get(headers.location as HttpURL, {
				'accept': 'application/ld+json',
				'user-agent': conf.user_agent,
			}));
		}

		if (status === 200) {
			const data = JSON.parse(body);
			return data as WeatherGovPoints;
		}

		const data = headers['content-type'].startsWith('application/problem+json')
			? JSON.parse(body)
			: { status };

		log.error(data, 'failed to fetch points data');
		throw new Error('failed to fetch points data');
	}
	
	async function get_forecast(gridId: string, gridX: number, gridY: number, units: 'us' | 'si', hourly = false) : Promise<WeatherGovForecast> {
		log.info({ gridId, gridX, gridY }, 'fetching forecast info');

		const url = `https://api.weather.gov/gridpoints/${gridId}/${gridX},${gridY}/forecast${hourly ? '/hourly' : ''}?units=${units}`;
		let { status, headers, body } = await outbound_http.get(url as HttpURL, {
			'accept': 'application/ld+json',
			'user-agent': conf.user_agent,
		});

		// Follow a single redirect if given
		if (status === 301) {
			({ status, headers, body } = await outbound_http.get(headers.location as HttpURL, {
				'accept': 'application/ld+json',
				'user-agent': conf.user_agent,
			}));
		}

		if (status === 200) {
			const data = JSON.parse(body);
			return data as WeatherGovForecast;
		}

		const data = headers['content-type'].startsWith('application/problem+json')
			? JSON.parse(body)
			: { status };

		log.error(data, 'failed to fetch forecast data');
		throw new Error('failed to fetch forecast data');
	}
	
	async function get_alerts(latitude: number, longitude: number) : Promise<WeatherGovAlerts> {
		log.info({ latitude, longitude }, 'fetching alerts info');

		let { status, headers, body } = await outbound_http.get(`https://api.weather.gov/alerts?active=1&point=${latitude},${longitude}`, {
			'accept': 'application/ld+json',
			'user-agent': conf.user_agent,
		});

		// Follow a single redirect if given
		if (status === 301) {
			({ status, headers, body } = await outbound_http.get(headers.location as HttpURL, {
				'accept': 'application/ld+json',
				'user-agent': conf.user_agent,
			}));
		}

		if (status === 200) {
			const data = JSON.parse(body);
			return data as WeatherGovAlerts;
		}

		const data = headers['content-type'].startsWith('application/problem+json')
			? JSON.parse(body)
			: { status };

		log.error(data, 'failed to fetch alerts data');
		throw new Error('failed to fetch alerts data');
	}
}
