
import pino from 'pino';
import { render_current_weather } from './render-current-weather';
import { SecretValue, resolve_secret_value } from '../../conf';
import type { NamedLocation } from '../../storage/named-location';
import type { OutboundHttpProvider } from '../../http/outbound';
import type { WeatherAPIComCurrentWeather } from './interface';

export interface WeatherAPIComConfig {
	enabled: boolean;
	api_key: SecretValue;
}

export function validate_weatherapi_com_conf(conf: unknown) : asserts conf is WeatherAPIComConfig {
	// todo: validate config
}

export type WeatherAPIComProvider = ReturnType<typeof create_weatherapi_com_provider>;

export function create_weatherapi_com_provider(conf: WeatherAPIComConfig, logger: pino.Logger, outbound_http: OutboundHttpProvider) {
	const log = logger.child({ logger: 'weatherapi.com' });
	const api_key = resolve_secret_value(conf.api_key);

	return {
		ready: Promise.resolve(),

		get enabled() {
			return conf?.enabled;
		},

		get_current: if_enabled(get_current),
		get_current_for_location(location: NamedLocation) {
			return get_current(`${location.latitude},${location.longitude}`);
		},

		render_current_weather,
	};

	function if_enabled<T>(func: T) : T {
		if (conf?.enabled) {
			return func;
		}

		return (() => {
			throw new Error('weatherapi.com service not enabled');
		}) as T;
	}
	
	async function get_current(query: string) : Promise<WeatherAPIComCurrentWeather> {
		log.info({ query }, 'fetching current weather');

		let { status, headers, body } = await outbound_http.get(`https://api.weatherapi.com/v1/current.json?key=${api_key}&q=${query}`, {
			'accept': 'application/json',
		});

		if (status === 200) {
			const data = JSON.parse(body);
			return data as WeatherAPIComCurrentWeather;
		}

		const data = headers['content-type'].startsWith('application/json')
			? JSON.parse(body)
			: { status };

		log.error(data, 'failed to fetch current weather');
		throw new Error('failed to fetch current weather');
	}
}
