
import { load_conf, validate_conf } from './conf';

import { create_logger } from './logger';
import { create_oidc_provider } from './security/openid-connect';
import { create_pkce_cookie_provider } from './security/pkce-cookie';
import { create_snowflake_provider } from './utilities/snowflake-uid';
import { create_storage_provider } from './storage';
import { create_argon_hash_provider } from './security/argon-hash';
import { create_session_provider } from './security/session';

import { create_http_metadata_server } from './http-metadata/server';
import { HttpWebDependencies, create_http_web_server } from './http-web/server';
import { create_color_theme_provider } from './utilities/color-themes';
import { create_outbound_http_provider } from './http/outbound';
import { create_services_provider } from './services';
import { create_rss_feed_reader_provider } from './utilities/rss-feeds';

main();

async function main() {
	// Load and validate configuration
	const conf = await load_conf();
	validate_conf(conf);
	
	// Create the logger and storage
	const logger = create_logger(conf.logging);
	const storage = create_storage_provider(conf.storage, logger.child({ logger: 'storage' }));

	// Create the metadata server
	const http_meta = create_http_metadata_server(conf.http_meta, {
		logger: logger.child({ logger: 'http-meta' }),
		ready_check,
		storage,
	});
	await http_meta.listen();

	// Create all of the core feature providers
	const oidc = create_oidc_provider(conf.oidc, logger);
	const pkce_cookie = create_pkce_cookie_provider(conf.pkce_cookie);
	const snowflake = create_snowflake_provider(conf.snowflake_uid);
	const argon2 = create_argon_hash_provider(conf.argon2);
	const session = create_session_provider(conf.session_cookie, logger, argon2, storage);
	const color_themes = create_color_theme_provider(conf.color_themes);
	const outbound_http = create_outbound_http_provider(conf.outbound_http, logger);
	const rss_feed_reader = create_rss_feed_reader_provider(logger, outbound_http);
	const services = create_services_provider(conf.services, logger, outbound_http);

	// Wait for any async init steps
	await oidc.ready;
	await storage.ready;
	await color_themes.ready;
	await services.ready;

	// Perform any cleanup steps before starting up
	await storage.cleanup_old_sessions();

	const deps: HttpWebDependencies = {
		logger: logger.child({ logger: 'http-web' }),
		oidc,
		pkce_cookie,
		snowflake,
		storage,
		argon2,
		session,
		color_themes,
		outbound_http,
		rss_feed_reader,
		services,
	};

	// Create the main web server
	const http_web = create_http_web_server(conf.http_web, deps);
	await http_web.setup_plugins();
	await http_web.listen();

	function ready_check() {
		try {
			const status = http_web.status;
			return status;
		}

		catch (error) {
			return 'unstarted';
		}
	}
}
