
import pino from 'pino';
import type { HttpURL } from './types';
import type { OutboundHttpProvider } from '../http/outbound';
import { XMLParser, XMLValidator, X2jOptions } from 'fast-xml-parser';

export type RSSFeedReaderProvider = ReturnType<typeof create_rss_feed_reader_provider>;

export function create_rss_feed_reader_provider(logger: pino.Logger, outbound_http: OutboundHttpProvider) {
	const log = logger.child({ logger: 'rss-feed-reader' });

	return {
		async read_feed(feed_url: HttpURL, skip_cache = false) {
			const req_headers = {
				accept: 'application/rss+xml; q=1.0, text/xml; q=0.9',
			};

			const { status, headers, body } = await outbound_http.get(feed_url, req_headers, skip_cache);

			if (status === 200) {
				const content_type = headers['content-type'] || '';

				if (content_type.startsWith('text/xml') || content_type.startsWith('application/rss+xml')) {
					const opts: Partial<X2jOptions> = { };
					const result = XMLValidator.validate(body, opts);
				
					if (typeof result === 'object' && result.err) {
						log.error(result.err, 'RSS feed XML validation error');
						throw_error('received invalid XML response from server');
					}
				
					const xmlParser = new XMLParser(opts);
					return xmlParser.parse(body);
				}

				else {
					throw_error('received invalid content type from server');
				}
			}

			else {
				throw_error('received a non-200 response code');
			}
		},
	};

	function throw_error(message: string) : never {
		throw new Error(`failed to fetch RSS feed content; ${message}`);
	}
}
